// PassStronger

#include <BleKeyboard.h>
#include <M5StickC.h>
#include <WiFi.h>
#include "Parameters.h"
#include "Bitmaps.h"
#include "EEPROM.h"
#include <hwcrypto/aes.h>

// Get pincode or default
uint16_t getpin(){
  uint16_t pincode=0;
  int i;
  if(ISSET){
    for (i=0;i<BLEPIN;++i)
      pincode=(EEPROM.readChar(2+PADLEN+i)-'0')+pincode*10;}
  return(pincode);
}

int lighton=0; // Set to 1 to avoid auto shut off
uint16_t pinvar=0;
BleKeyboard bleKeyboard(DEVICE_NAME,"PassStrong2",99,HID_TYPE,&pinvar); // The last parameter only work with a modified version of Blekeyboard library. Discard it if you don't have it.
int refresh=3000/DELAYL;// Main loop Counter 
int dmode=-1;           // Mode -1:Try to connect 0: Main Menu 1,2,3,4,5: Passn choice menu
int canwrite=0;         // Pin write mode ? 
char pin[CODELEN+2];    // Pin code
char opin[CODELEN+2];   // Pin backup for code change
int passn=0;            // Mode qwerty or azerty
int confgen=0;          // Regen confirmation mode

// Encoding routines taken from CNLohr
// You must be encrypting/decrypting data that is in BLOCKSIZE chunks!!!
char cryptpass[PADLEN];

// Genkey : for added security, you need to change AFACTOR and BFACTOR for each unit, or one can guess the key, and uncrypt the password
void genkey(char *p){
  uint8_t j=0,r=1;
  for (int i=0;i<KSIZE;++i){
    r=(p[j]*AFACTOR+BFACTOR*r)&255;
    key[i]=r;
    j=(j+1)%CODELEN;}
}



// Clear password
void clearpass(){
  memset((uint8_t*) password, 0, PADLEN ); // Clear 
}

// Crypts password into EEPROM and clears password
void encrypt(char *p=pin){
  for (int i=0;i<CODELEN;++i) password[PASSLEN+i]=p[i];
  password[PASSLEN+CODELEN]='O';  
  password[PASSLEN+CODELEN+1]='K';  
  for (int i=PASSLEN+CODELEN+2;i<PADLEN;) password[i++]=esp_random();
  esp_aes_context ctx;
  memset( iv, 0, sizeof( iv ) ); 
  esp_aes_init( &ctx );
  genkey(p);
  esp_aes_setkey( &ctx, key, KSIZE*8 );
  esp_aes_crypt_cbc( &ctx, ESP_AES_ENCRYPT, PADLEN, iv, (uint8_t*)password, (uint8_t*)cryptpass );
  for (int i=0;i<PADLEN;++i) EEPROM.writeChar(2+i,cryptpass[i]);
  EEPROM.commit();
  clearpass();
  esp_aes_free( &ctx );
}

// Uncrypts password from EEPROM. Nb : do clearpass() as soon as possible 
void decrypt(char *p=pin){
  esp_aes_context ctx;
  memset( iv, 0, sizeof( iv ) );  
  esp_aes_init( &ctx );
  for (int i=0;i<PADLEN;++i) cryptpass[i]=EEPROM.readChar(2+i);
  genkey(p);
  esp_aes_setkey( &ctx, key, KSIZE*8 );
  esp_aes_crypt_cbc( &ctx, ESP_AES_DECRYPT, PADLEN, iv, (uint8_t*)cryptpass, (uint8_t*)password );
  esp_aes_free( &ctx );
}

// Regenerate random password and flashit
void regen(){
  for (int i=0;i<PASSLEN;++i) 
      password[i]=possletters[esp_random()%possletters.length()];
      #ifdef DEBUG
        password[PASSLEN]=0;
        Serial.println("Regen");
        Serial.println(password);
      #endif
  encrypt();
}

// Check if pin is correct
bool pinok(char *p=pin){
  decrypt(p);
  return((password[PASSLEN+CODELEN]=='O')&&(password[PASSLEN+CODELEN+1]=='K'));
}

// Qwerty printer
void qwprint(String text){
 for (int i=0;i<PASSLEN;++i){
   bleKeyboard.press(text[i]&127);
   delay(50);
   bleKeyboard.releaseAll();}
}

// Azerty printer
void azkprint(String text){
 for (int i=0;i<PASSLEN;++i){
   uint16_t c=qwert_conv[(unsigned int)(text[i]&127)];
   if (c&SHIFT) bleKeyboard.press(KEY_LEFT_SHIFT);
   if (c&ALTGR) bleKeyboard.press(KEY_RIGHT_ALT);
   bleKeyboard.press(c&127);
   delay(50);
   bleKeyboard.releaseAll();}
}

// Shows bitmap
void showbit(const unsigned char *btm){
  M5.Lcd.fillScreen(TFT_BLACK); // Black screen fill
  M5.Lcd.drawXBitmap(0,0, btm, 80, 160, TFT_WHITE);}

// Types password
void bttype(){
  bleKeyboard.releaseAll();
  delay(DELUL*4);
  showbit(sending);
  #ifdef DEBUG
   Serial.println(modest[passn]);
  #endif
  if (passn) azkprint(password);
  else qwprint(password);
  delay(DELUL*4);}

  
// Device setup
void setup(void) {
  int i;
  while (!EEPROM.begin(EEPROM_SIZE)) true;
  M5.begin();
  esp_sleep_enable_ext0_wakeup(GPIO_NUM_39, 0);
  WiFi.mode(WIFI_MODE_NULL); // Shut off wifi
  pinMode(37,INPUT);
  pinMode(BUTTON_B_PIN,INPUT);
  pinMode(BUTTON_A_PIN,INPUT);
  while (digitalRead(BUTTON_B_PIN)==LOW){ // Wipe code
    for (i=0;i<2+PADLEN+BLEPIN;++i) EEPROM.writeChar(i,0xFF);
    EEPROM.commit();
    delay(100);}
  showbit(init1);
#ifdef DEBUG
  Serial.begin(115200);
  Serial.print("Starting ");
  Serial.println(getpin());
#endif
  for (int i=0;i<CODELEN+1;++i) pin[i]='0';
  pin[CODELEN]='\0';
  if (!ISSET){
   regen(); // If blank device, generate a password to avoid problem with random non ascii characters
   int i;
   for (i=0;i<BLEPIN;++i)
      EEPROM.writeChar(2+PADLEN+i,'0');
   EEPROM.commit();
  }
  pinvar=getpin();
  bleKeyboard.begin();
}

// Settings status
void setstat(){
  int j;
  char buf[CODELEN+1];
  char buf2[BLEPIN+2];
  M5.Lcd.fillScreen(TFT_BLACK);
  M5.Lcd.setTextColor(TFT_WHITE);  // Adding a background colour erases previous text automatically
  M5.Lcd.setTextSize(2);
  M5.Lcd.setCursor(0,18);  
  M5.Lcd.print(String(pin));
  if ((dmode>0)&&(dmode<5)){
    for (j=0;j<dmode-1;) buf[j++]=' ';
    buf[j]='^';
    buf[j+1]=0;  
    M5.Lcd.setCursor(0,36);  
    M5.Lcd.print(String(buf));}  
  M5.Lcd.setCursor(0,54);
  M5.Lcd.print(dmode==5?">Chnge":"Change");  
  M5.Lcd.setCursor(0,72);  
  M5.Lcd.print(dmode==6?">Mode:":"Mode:");  
  M5.Lcd.setCursor(0,90);  
  M5.Lcd.print(modest[passn]);  
  M5.Lcd.setCursor(0,108);  
  M5.Lcd.print(dmode==7?(confgen==1?"*Regen":">Regen"):"Regen");  
  M5.Lcd.setCursor(0,126);  
  M5.Lcd.print(ISSET?"Pinset":"No pin");  
  M5.Lcd.setCursor(0,144); 
  for (j=0;j<BLEPIN;++j) buf2[j+1]=EEPROM.readChar(2+PADLEN+j);
  buf2[BLEPIN+1]=0;
  buf2[0]='>';
  M5.Lcd.print(String((dmode==8)?buf2:(buf2+1)));}

// Action if small button pressed
void dosmall(){
  switch(dmode){
    case 0: case -1:// Home menu
       if (pinok()){
         bttype();
         dmode=-1;
         confgen=0;
         refresh=80;}
       clearpass(); // Needed after pinok to clear typed password from memory
       break;
    case 8: // New blepin
       int j;
       M5.Lcd.setCursor(0,144);  
       M5.Lcd.print("*****");
       if (EEPROM.readChar(2+PADLEN)%3==1){
        for (j=0;j<BLEPIN;++j)
           EEPROM.writeChar(2+PADLEN+j,'0');
        }
       else
        for (j=0;j<BLEPIN;++j)
           EEPROM.writeChar(2+PADLEN+j,(esp_random()%(j==0?9:10))+(j==0?'1':'0'));
       EEPROM.commit();
       delay(100);
       break;
       
    case 7: // Regen menu
        if (confgen==0) confgen=1; // Ask for confirmation
        else{    
          confgen=0;               // Confirmed
          if (pinok()){
           bttype();
           regen();
           M5.Lcd.setCursor(0,108);  
           M5.Lcd.print("#");
           delay(300);
           refresh=80;
           dmode=-1;}
          clearpass();}
        break;
    case 6: 
        canwrite=0;
        passn=(passn+1)%2; // AZ/QW ERTY shift
        break;
    case 5: // Change pin
    #ifdef DEBUG
        Serial.println(ISSET?"ON":"OFF");
        Serial.println(pin);
        Serial.println(opin);
    #endif
        if (canwrite or !ISSET){  // Is it possible (either entered correct pin, or no pin set)
          canwrite=0;
          M5.Lcd.setCursor(0,54);  
          M5.Lcd.print("#");
          delay(300);
          if (pinok(opin)||(!ISSET)){
            DOSET;
            encrypt(); // Re encrypt with new pin
            dmode+=1;}
          clearpass();}
        else{  // First pass on the menu
          if (pinok()){ // If pin is correct then back to pin to enter the new one
           dmode=1;
           canwrite=1;
           M5.Lcd.setCursor(0,18);  
           M5.Lcd.print("####");
           delay(300);}
          clearpass();  
        }
        break;
    case 1: case 2: case 3: case 4:
        pin[dmode-1]=pin[dmode-1]>='9'?'0':pin[dmode-1]+1;
        refresh=2*300/DELAYL;
        break;
  }
  if (dmode>0) setstat();
}

void dobig(){
 dmode=(dmode+1)%9;
 if (dmode==0) showbit(menu);
 else {
   
   if (dmode==1) 
      {for (int i=0;i<CODELEN+1;++i) opin[i]=pin[i];} // Save old pin in case it is changed
    setstat();}
 refresh=20;
}

void loop() {
 if (dmode==0){
  if ((lighton==0)&&(millis()>shut_time)){
    M5.Axp.DeepSleep();}}
 else
    shut_time=MILLILIGHT+millis();   
    
 refresh-=2;
 if (refresh<=0){
  refresh+=11;
  if (bleKeyboard.isConnected()){
    if (dmode<0){
      if (dmode==-2){
      showbit(connectok);
      delay(500);}
      showbit(menu);
      dmode=0;
      }
    if (digitalRead(BUTTON_B_PIN)==LOW){
      if (!lighton)
        shut_time=MILLILIGHT+millis();
      dosmall();}
   else if (digitalRead(BUTTON_A_PIN)==LOW){
      if (!lighton)
        shut_time=MILLILIGHT+millis();   
        dobig();}
   }
  else {
    showbit(refresh&1?init1:init2);
    dmode=-2;
    }
  }
 delay(DELAYL);
}
